'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

// Modified version of S.js[https://github.com/adamhaile/S] by Adam Haile
// Public interface
function createSignal(value, comparator) {
  const d = new DataNode(value);
  let setter;

  if (comparator) {
    let age = -1;

    setter = v => {
      if (!comparator(value, v)) {
        const time = RootClock.time;

        if (time === age) {
          throw new Error(`Conflicting value update: ${v} is not the same as ${value}`);
        }

        age = time;
        value = v;
        d.next(v);
      }
    };
  } else setter = d.next.bind(d);

  return [d.current.bind(d), setter];
}
function createEffect(fn, value) {
  makeComputationNode(fn, value, false);
} // explicit dependencies and defered initial execution

function createDependentEffect(fn, deps, defer) {
  if (Array.isArray(deps)) deps = callAll(deps);
  defer = !!defer;
  createEffect(value => {
    const listener = Listener;
    deps();
    if (defer) defer = false;else {
      Listener = null;
      value = fn(value);
      Listener = listener;
    }
    return value;
  });
}
function createMemo(fn, value) {
  var {
    node,
    value: _value
  } = makeComputationNode(fn, value, false);
  return node === null ? () => _value : () => {
    if (Listener !== null) {
      if (node.age === RootClock.time) {
        if (node.state === RUNNING) throw new Error("circular dependency");else updateNode(node); // checks for state === STALE internally, so don't need to check here
      }

      if (node.log === null) node.log = createLog();
      logRead(node.log);
    }

    return node.value;
  };
}
function createRoot(fn, detachedOwner) {
  let disposer = fn.length === 0 ? null : function _dispose() {
    if (root === null) ; else if (RunningClock !== null) {
      RootClock.disposes.add(root);
    } else dispose(root);
  },
      root = disposer === null ? UNOWNED : getCandidateNode(),
      result,
      listener = Listener,
      owner = detachedOwner || Owner;
  root !== owner && (root.owner = owner);
  Owner = root;

  try {
    Listener = null;
    result = disposer === null ? fn() : fn(disposer);
  } finally {
    Listener = listener;
    Owner = root.owner;
  }

  if (disposer !== null && recycleOrClaimNode(root, null, undefined, true)) {
    root = null;
  }

  return result;
}
function freeze(fn) {
  let result = undefined;
  if (RunningClock !== null) result = fn();else {
    RunningClock = RootClock;
    RunningClock.changes.reset();

    try {
      result = fn();
      event();
    } finally {
      RunningClock = null;
    }
  }
  return result;
}
function sample(fn) {
  let result,
      listener = Listener;
  Listener = null;
  result = fn();
  Listener = listener;
  return result;
}
function onCleanup(fn) {
  if (Owner === null) console.warn("cleanups created without a root or parent will never be run");else if (Owner.cleanups === null) Owner.cleanups = [fn];else Owner.cleanups.push(fn);
}
function afterEffects(fn) {
  Promise.resolve().then(fn);
}
function isListening() {
  return Listener !== null;
} // context API

function createContext(initFn) {
  const id = Symbol("context");
  return {
    id,
    Provider: createProvider(id, initFn)
  };
}
function useContext(context) {
  if (Owner === null) return console.warn("Context keys cannot be looked up without a root or parent");
  return lookup(Owner, context.id);
}
function getContextOwner() {
  Owner && (Owner.noRecycle = true);
  return Owner;
} // Internal implementation
/// Graph classes and operations

class DataNode {
  constructor(value) {
    this.value = value;
    this.pending = NOTPENDING;
    this.log = null;
  }

  current() {
    if (Listener !== null) {
      if (this.log === null) this.log = createLog();
      logRead(this.log);
    }

    return this.value;
  }

  next(value) {
    if (RunningClock !== null) {
      if (this.pending !== NOTPENDING) {
        // value has already been set once, check for conflicts
        if (value !== this.pending) {
          throw new Error("conflicting changes: " + value + " !== " + this.pending);
        }
      } else {
        // add to list of changes
        this.pending = value;
        RootClock.changes.add(this);
      }
    } else {
      // not batching, respond to change now
      if (this.log !== null) {
        this.pending = value;
        RootClock.changes.add(this);
        event();
      } else {
        this.value = value;
      }
    }

    return value;
  }

}

function createComputationNode() {
  return {
    fn: null,
    age: -1,
    state: CURRENT,
    source1: null,
    source1slot: 0,
    sources: null,
    sourceslots: null,
    owner: null,
    owned: null,
    log: null,
    value: undefined,
    context: undefined,
    cleanups: null
  };
}

function createClock() {
  return {
    time: 0,
    changes: new Queue(),
    // batched changes to data nodes
    updates: new Queue(),
    // computations to update
    disposes: new Queue() // disposals to run after current batch of updates finishes

  };
}

function createLog() {
  return {
    node1: null,
    node1slot: 0,
    nodes: null,
    nodeslots: null
  };
}

class Queue {
  constructor() {
    this.items = [];
    this.count = 0;
  }

  reset() {
    this.count = 0;
  }

  add(item) {
    this.items[this.count++] = item;
  }

  run(fn) {
    let items = this.items;

    for (let i = 0; i < this.count; i++) {
      fn(items[i]);
      items[i] = null;
    }

    this.count = 0;
  }

}

function createProvider(id, initFn) {
  return props => {
    let rendered;
    makeComputationNode(() => {
      if (Owner === null) return console.warn("Context keys cannot be set without a root or parent");
      const context = Owner.context || (Owner.context = {});
      Owner.noRecycle = true;
      context[id] = initFn ? initFn(props.value) : props.value;
      rendered = props.children;
    }, undefined, true);
    return rendered;
  };
} // "Globals" used to keep track of current system state


let RootClock = createClock(),
    RunningClock = null,
    // currently running clock
Listener = null,
    // currently listening computation
Owner = null,
    // owner for new computations
LastNode = null; // cached unused node, for re-use
// Constants

let NOTPENDING = {},
    CURRENT = 0,
    STALE = 1,
    RUNNING = 2,
    UNOWNED = createComputationNode(); // Functions

function callAll(ss) {
  return function all() {
    for (let i = 0; i < ss.length; i++) ss[i]();
  };
}

function lookup(owner, key) {
  return owner && owner.context && owner.context[key] || owner.owner && lookup(owner.owner, key);
}

var makeComputationNodeResult = {
  node: null,
  value: undefined
};

function makeComputationNode(fn, value, sample) {
  const node = getCandidateNode(),
        listener = Listener,
        toplevel = RunningClock === null;
  node.owner = Owner;
  Owner = node;
  Listener = sample ? null : node;

  if (toplevel) {
    value = execToplevelComputation(fn, value);
  } else {
    value = fn(value);
  }

  Owner = node.owner;
  Listener = listener;
  var recycled = recycleOrClaimNode(node, fn, value, false);
  if (toplevel) finishToplevelComputation(Owner, listener);
  makeComputationNodeResult.node = recycled ? null : node;
  makeComputationNodeResult.value = value;
  return makeComputationNodeResult;
}

function execToplevelComputation(fn, value) {
  RunningClock = RootClock;
  RootClock.changes.reset();
  RootClock.updates.reset();

  try {
    return fn(value);
  } finally {
    Owner = Listener = RunningClock = null;
  }
}

function finishToplevelComputation(owner, listener) {
  if (RootClock.changes.count > 0 || RootClock.updates.count > 0) {
    RootClock.time++;

    try {
      run(RootClock);
    } finally {
      RunningClock = null;
      Owner = owner;
      Listener = listener;
    }
  }
}

function getCandidateNode() {
  let node = LastNode;
  if (node === null) node = createComputationNode();else LastNode = null;
  return node;
}

function recycleOrClaimNode(node, fn, value, orphan) {
  let _owner = orphan || Owner === null || Owner === UNOWNED ? null : Owner,
      recycle = !node.noRecycle && node.source1 === null && (node.owned === null && node.cleanups === null || _owner !== null),
      i;

  if (recycle) {
    LastNode = node;
    node.owner = null;

    if (_owner !== null) {
      if (node.owned !== null) {
        if (_owner.owned === null) _owner.owned = node.owned;else for (i = 0; i < node.owned.length; i++) {
          _owner.owned.push(node.owned[i]);
        }
        node.owned = null;
      }

      if (node.cleanups !== null) {
        if (_owner.cleanups === null) _owner.cleanups = node.cleanups;else for (i = 0; i < node.cleanups.length; i++) {
          _owner.cleanups.push(node.cleanups[i]);
        }
        node.cleanups = null;
      }
    }
  } else {
    node.fn = fn;
    node.value = value;
    node.age = RootClock.time;

    if (_owner !== null) {
      if (_owner.owned === null) _owner.owned = [node];else _owner.owned.push(node);
    }
  }

  return recycle;
}

function logRead(from) {
  let to = Listener,
      fromslot,
      toslot = to.source1 === null ? -1 : to.sources === null ? 0 : to.sources.length;

  if (from.node1 === null) {
    from.node1 = to;
    from.node1slot = toslot;
    fromslot = -1;
  } else if (from.nodes === null) {
    if (from.node1 === to) return;
    from.nodes = [to];
    from.nodeslots = [toslot];
    fromslot = 0;
  } else {
    fromslot = from.nodes.length;
    if (from.nodes[fromslot - 1] === to) return;
    from.nodes.push(to);
    from.nodeslots.push(toslot);
  }

  if (to.source1 === null) {
    to.source1 = from;
    to.source1slot = fromslot;
  } else if (to.sources === null) {
    to.sources = [from];
    to.sourceslots = [fromslot];
  } else {
    to.sources.push(from);
    to.sourceslots.push(fromslot);
  }
}

function event() {
  // b/c we might be under a top level S.root(), have to preserve current root
  let owner = Owner;
  RootClock.updates.reset();
  RootClock.time++;

  try {
    run(RootClock);
  } finally {
    RunningClock = Listener = null;
    Owner = owner;
  }
}

function run(clock) {
  let running = RunningClock,
      count = 0;
  RunningClock = clock;
  clock.disposes.reset(); // for each batch ...

  while (clock.changes.count !== 0 || clock.updates.count !== 0 || clock.disposes.count !== 0) {
    if (count > 0) // don't tick on first run, or else we expire already scheduled updates
      clock.time++;
    clock.changes.run(applyDataChange);
    clock.updates.run(updateNode);
    clock.disposes.run(dispose); // if there are still changes after excessive batches, assume runaway

    if (count++ > 1e5) {
      throw new Error("Runaway clock detected");
    }
  }

  RunningClock = running;
}

function applyDataChange(data) {
  data.value = data.pending;
  data.pending = NOTPENDING;
  if (data.log) markComputationsStale(data.log);
}

function markComputationsStale(log) {
  let node1 = log.node1,
      nodes = log.nodes; // mark all downstream nodes stale which haven't been already

  if (node1 !== null) markNodeStale(node1);

  if (nodes !== null) {
    for (let i = 0, len = nodes.length; i < len; i++) {
      markNodeStale(nodes[i]);
    }
  }
}

function markNodeStale(node) {
  let time = RootClock.time;

  if (node.age < time) {
    node.age = time;
    node.state = STALE;
    RootClock.updates.add(node);
    if (node.owned !== null) markOwnedNodesForDisposal(node.owned);
    if (node.log !== null) markComputationsStale(node.log);
  }
}

function markOwnedNodesForDisposal(owned) {
  for (let i = 0; i < owned.length; i++) {
    let child = owned[i];
    child.age = RootClock.time;
    child.state = CURRENT;
    if (child.owned !== null) markOwnedNodesForDisposal(child.owned);
  }
}

function updateNode(node) {
  if (node.state === STALE) {
    let owner = Owner,
        listener = Listener;
    Owner = Listener = node;
    node.state = RUNNING;
    cleanupNode(node, false);
    node.value = node.fn(node.value);
    node.state = CURRENT;
    Owner = owner;
    Listener = listener;
  }
}

function cleanupNode(node, final) {
  let source1 = node.source1,
      sources = node.sources,
      sourceslots = node.sourceslots,
      cleanups = node.cleanups,
      owned = node.owned,
      i,
      len;

  if (cleanups !== null) {
    for (i = 0; i < cleanups.length; i++) {
      cleanups[i](final);
    }

    node.cleanups = null;
  }

  if (owned !== null) {
    for (i = 0; i < owned.length; i++) {
      dispose(owned[i]);
    }

    node.owned = null;
  }

  if (source1 !== null) {
    cleanupSource(source1, node.source1slot);
    node.source1 = null;
  }

  if (sources !== null) {
    for (i = 0, len = sources.length; i < len; i++) {
      cleanupSource(sources.pop(), sourceslots.pop());
    }
  }
}

function cleanupSource(source, slot) {
  let nodes = source.nodes,
      nodeslots = source.nodeslots,
      last,
      lastslot;

  if (slot === -1) {
    source.node1 = null;
  } else {
    last = nodes.pop();
    lastslot = nodeslots.pop();

    if (slot !== nodes.length) {
      nodes[slot] = last;
      nodeslots[slot] = lastslot;

      if (lastslot === -1) {
        last.source1slot = slot;
      } else {
        last.sourceslots[lastslot] = slot;
      }
    }
  }
}

function dispose(node) {
  node.fn = null;
  node.owner = null;
  node.log = null;
  cleanupNode(node, true);
}

const SNODE = Symbol("solid-node"),
      SPROXY = Symbol("solid-proxy");

function wrap(value) {
  return value[SPROXY] || (value[SPROXY] = new Proxy(value, proxyTraps));
}

function isWrappable(obj) {
  return obj !== null && typeof obj === "object" && (obj.__proto__ === Object.prototype || Array.isArray(obj));
}
function unwrap(item) {
  let result, unwrapped, v;
  if (result = item != null && item._state) return result;
  if (!isWrappable(item)) return item;

  if (Array.isArray(item)) {
    if (Object.isFrozen(item)) item = item.slice(0);

    for (let i = 0, l = item.length; i < l; i++) {
      v = item[i];
      if ((unwrapped = unwrap(v)) !== v) item[i] = unwrapped;
    }
  } else {
    if (Object.isFrozen(item)) item = Object.assign({}, item);
    let keys = Object.keys(item);

    for (let i = 0, l = keys.length; i < l; i++) {
      v = item[keys[i]];
      if ((unwrapped = unwrap(v)) !== v) item[keys[i]] = unwrapped;
    }
  }

  return item;
}

function getDataNodes(target) {
  let nodes = target[SNODE];
  if (!nodes) target[SNODE] = nodes = {};
  return nodes;
}

const proxyTraps = {
  get(target, property) {
    if (property === "_state") return target;
    if (property === SPROXY || property === SNODE) return;
    const value = target[property],
          wrappable = isWrappable(value);

    if (isListening() && typeof value !== "function") {
      let nodes, node;

      if (wrappable && (nodes = getDataNodes(value))) {
        node = nodes._ || (nodes._ = new DataNode());
        node.current();
      }

      nodes = getDataNodes(target);
      node = nodes[property] || (nodes[property] = new DataNode());
      node.current();
    }

    return wrappable ? wrap(value) : value;
  },

  set() {
    return true;
  },

  deleteProperty() {
    return true;
  }

};
function setProperty(state, property, value) {
  value = unwrap(value);
  if (state[property] === value) return;
  const notify = Array.isArray(state) || !(property in state);

  if (value === void 0) {
    delete state[property];
  } else state[property] = value;

  let nodes = getDataNodes(state),
      node;
  (node = nodes[property]) && node.next();
  notify && (node = nodes._) && node.next();
}

function mergeState(state, value) {
  const keys = Object.keys(value);

  for (let i = 0; i < keys.length; i += 1) {
    const key = keys[i];
    setProperty(state, key, value[key]);
  }
}

function updatePath(current, path, traversed = []) {
  if (path.length === 1) {
    let value = path[0];

    if (typeof value === "function") {
      value = value(wrap(current), traversed); // reconciled

      if (value === undefined) return;
    }

    mergeState(current, value);
    return;
  }

  const part = path.shift(),
        partType = typeof part,
        isArray = Array.isArray(current);

  if (Array.isArray(part)) {
    // Ex. update('data', [2, 23], 'label', l => l + ' !!!');
    for (let i = 0; i < part.length; i++) {
      updatePath(current, [part[i]].concat(path), traversed.concat([part[i]]));
    }
  } else if (isArray && partType === "function") {
    // Ex. update('data', i => i.id === 42, 'label', l => l + ' !!!');
    for (let i = 0; i < current.length; i++) {
      if (part(current[i], i)) updatePath(current, [i].concat(path), traversed.concat([i]));
    }
  } else if (isArray && partType === "object") {
    // Ex. update('data', { from: 3, to: 12, by: 2 }, 'label', l => l + ' !!!');
    const {
      from = 0,
      to = current.length - 1,
      by = 1
    } = part;

    for (let i = from; i <= to; i += by) {
      updatePath(current, [i].concat(path), traversed.concat([i]));
    }
  } else if (path.length === 1) {
    let value = path[0];

    if (typeof value === "function") {
      const currentPart = current[part];
      value = value(isWrappable(currentPart) ? wrap(currentPart) : currentPart, traversed.concat([part]));
    }

    if (isWrappable(current[part]) && isWrappable(value) && !Array.isArray(value)) {
      mergeState(current[part], value);
    } else setProperty(current, part, value);
  } else updatePath(current[part], path, traversed.concat([part]));
}

function createState(state) {
  state = unwrap(state || {});
  const wrappedState = wrap(state);

  function setState(...args) {
    freeze(() => {
      if (Array.isArray(args[0])) {
        for (let i = 0; i < args.length; i += 1) {
          updatePath(state, args[i]);
        }
      } else updatePath(state, args);
    });
  }

  return [wrappedState, setState];
}

function applyState(target, parent, property, merge, key) {
  let previous = parent[property];
  if (target === previous) return;

  if (!isWrappable(target) || previous == null) {
    target !== previous && setProperty(parent, property, target);
    return;
  }

  if (Array.isArray(target)) {
    if (target.length && previous.length && (!merge || key && target[0][key] != null)) {
      let i, j, start, end, newEnd, item, newIndicesNext, keyVal; // common prefix

      for (start = 0, end = Math.min(previous.length, target.length); start < end && (previous[start] === target[start] || key && previous[start][key] === target[start][key]); start++) {
        applyState(target[start], previous, start, merge, key);
      }

      const temp = new Array(target.length),
            newIndices = new Map(); // common suffix

      for (end = previous.length - 1, newEnd = target.length - 1; end >= start && newEnd >= start && (previous[end] === target[newEnd] || key && previous[end][key] === target[newEnd][key]); end--, newEnd--) {
        temp[newEnd] = previous[end];
      } // remove any remaining nodes and we're done


      if (start > newEnd) {
        const rLen = end - start + 1;
        if (rLen > 0) previous.splice(start, rLen);
        setProperty(previous, "length", target.length);
        return;
      } // insert any remaining updates and we're done


      if (start > end) {
        for (j = start; j <= newEnd; j++) setProperty(previous, j, target[j]);

        for (; j < target.length; j++) {
          setProperty(previous, j, temp[j]);
          applyState(target[j], previous, j, merge, key);
        }

        return;
      } // prepare a map of all indices in target


      newIndicesNext = new Array(newEnd + 1);

      for (j = newEnd; j >= start; j--) {
        item = target[j];
        keyVal = key ? item[key] : item;
        i = newIndices.get(keyVal);
        newIndicesNext[j] = i === undefined ? -1 : i;
        newIndices.set(keyVal, j);
      } // step through all old items to check reuse


      for (i = start; i <= end; i++) {
        item = previous[i];
        keyVal = key ? item[key] : item;
        j = newIndices.get(keyVal);

        if (j !== undefined && j !== -1) {
          temp[j] = previous[i];
          j = newIndicesNext[j];
          newIndices.set(keyVal, j);
        }
      } // set all the new values


      for (j = start; j < target.length; j++) {
        if (j in temp) {
          setProperty(previous, j, temp[j]);
          applyState(target[j], previous, j, merge, key);
        } else setProperty(previous, j, target[j]);
      }
    } else {
      for (let i = 0, len = target.length; i < len; i++) {
        applyState(target[i], previous, i, merge, key);
      }
    }

    if (previous.length > target.length) setProperty(previous, "length", target.length);
    return;
  }

  const targetKeys = Object.keys(target);

  for (let i = 0, len = targetKeys.length; i < len; i++) {
    applyState(target[targetKeys[i]], previous, targetKeys[i], merge, key);
  }

  const previousKeys = Object.keys(previous);

  for (let i = 0, len = previousKeys.length; i < len; i++) {
    if (target[previousKeys[i]] === undefined) setProperty(previous, previousKeys[i], undefined);
  }
} // Diff method for setState


function reconcile(path, options = {}) {
  let value;

  if (Array.isArray(path)) {
    value = path.pop();
  } else if (typeof path === "object") {
    value = path;
    path = undefined;
  } else {
    path = Array.prototype.slice.call(arguments, 0, -1), value = arguments[arguments.length - 1];
    options = {};
  }

  const {
    merge,
    key = "id"
  } = options;
  return state => {
    state = unwrap(state);

    if (path) {
      for (let i = 0; i < path.length - 1; i += 1) state = state[path[i]];

      applyState(value, state, path[path.length - 1], merge, key);
    } else applyState(value, {
      state
    }, "state", merge, key);
  };
}

const FALLBACK = Symbol("fallback");
function pipe(...fns) {
  if (!fns) return i => i;
  if (fns.length === 1) return fns[0];
  return input => fns.reduce((prev, fn) => fn(prev), input);
} // Modified version of mapSample from S-array[https://github.com/adamhaile/S-array] by Adam Haile

function map(mapFn, fallback) {
  return list => {
    let items = [],
        mapped = [],
        disposers = [],
        len = 0;
    onCleanup(() => {
      for (let i = 0, length = disposers.length; i < length; i++) disposers[i]();
    });
    return () => {
      let newItems = list() || [],
          i,
          j;
      return sample(() => {
        let newLen = newItems.length,
            newIndices,
            newIndicesNext,
            temp,
            tempdisposers,
            start,
            end,
            newEnd,
            item; // fast path for empty arrays

        if (newLen === 0) {
          if (len !== 0) {
            for (i = 0; i < len; i++) disposers[i]();

            disposers = [];
            items = [];
            mapped = [];
            len = 0;
          }

          if (fallback) {
            items = [FALLBACK];
            mapped[0] = createRoot(disposer => {
              disposers[0] = disposer;
              return fallback();
            });
            len = 1;
          }
        } // fast path for new create
        else if (len === 0) {
            for (j = 0; j < newLen; j++) {
              items[j] = newItems[j];
              mapped[j] = createRoot(mapper);
            }

            len = newLen;
          } else {
            temp = new Array(newLen);
            tempdisposers = new Array(newLen); // skip common prefix

            for (start = 0, end = Math.min(len, newLen); start < end && items[start] === newItems[start]; start++); // common suffix


            for (end = len - 1, newEnd = newLen - 1; end >= start && newEnd >= start && items[end] === newItems[newEnd]; end--, newEnd--) {
              temp[newEnd] = mapped[end];
              tempdisposers[newEnd] = disposers[end];
            } // remove any remaining nodes and we're done


            if (start > newEnd) {
              for (j = end; start <= j; j--) disposers[j]();

              const rLen = end - start + 1;

              if (rLen > 0) {
                mapped.splice(start, rLen);
                disposers.splice(start, rLen);
              }

              items = newItems.slice(0);
              len = newLen;
              return mapped;
            } // insert any remaining updates and we're done


            if (start > end) {
              for (j = start; j <= newEnd; j++) mapped[j] = createRoot(mapper);

              for (; j < newLen; j++) {
                mapped[j] = temp[j];
                disposers[j] = tempdisposers[j];
              }

              items = newItems.slice(0);
              len = newLen;
              return mapped;
            } // 0) prepare a map of all indices in newItems, scanning backwards so we encounter them in natural order


            newIndices = new Map();
            newIndicesNext = new Array(newEnd + 1);

            for (j = newEnd; j >= start; j--) {
              item = newItems[j];
              i = newIndices.get(item);
              newIndicesNext[j] = i === undefined ? -1 : i;
              newIndices.set(item, j);
            } // 1) step through all old items and see if they can be found in the new set; if so, save them in a temp array and mark them moved; if not, exit them


            for (i = start; i <= end; i++) {
              item = items[i];
              j = newIndices.get(item);

              if (j !== undefined && j !== -1) {
                temp[j] = mapped[i];
                tempdisposers[j] = disposers[i];
                j = newIndicesNext[j];
                newIndices.set(item, j);
              } else disposers[i]();
            } // 2) set all the new values, pulling from the temp array if copied, otherwise entering the new value


            for (j = start; j < newLen; j++) {
              if (j in temp) {
                mapped[j] = temp[j];
                disposers[j] = tempdisposers[j];
              } else mapped[j] = createRoot(mapper);
            } // 3) in case the new set is shorter than the old, set the length of the mapped array


            len = mapped.length = newLen; // 4) save a copy of the mapped items for the next update

            items = newItems.slice(0);
          }

        return mapped;
      });

      function mapper(disposer) {
        disposers[j] = disposer;
        return mapFn(newItems[j], j);
      }
    };
  };
}
function reduce(fn, seed) {
  return list => () => {
    let newList = list() || [],
        result = seed;
    return sample(() => {
      for (let i = 0; i < newList.length; i++) {
        result = fn(result, newList[i], i);
      }

      return result;
    });
  };
}

function setDefaults(props, defaultProps) {
  const propKeys = Object.keys(defaultProps);

  for (let i = 0; i < propKeys.length; i++) {
    const key = propKeys[i];
    !(key in props) && (props[key] = defaultProps[key]);
  }
}

function createHandler(className) {
  return (e, s) => e.classList.toggle(className, s);
}

function shallowDiff(a, b) {
  let sa = new Set(a),
      sb = new Set(b);
  return [a.filter(i => !sb.has(i)), b.filter(i => !sa.has(i))];
}

function selectWhen(signal, handler) {
  if (typeof handler === "string") handler = createHandler(handler);
  return list => {
    createEffect(element => {
      const model = signal();
      if (element) handler(element, false);
      if (element = model && sample(list).find(el => el.model === model)) handler(element, true);
      return element;
    });
    return list;
  };
}
function selectEach(signal, handler) {
  if (typeof handler === "string") handler = createHandler(handler);
  return list => {
    createEffect((elements = []) => {
      const models = signal(),
            newElements = sample(list).filter(el => models.indexOf(el.model) > -1),
            [additions, removals] = shallowDiff(newElements, elements);
      additions.forEach(el => handler(el, true));
      removals.forEach(el => handler(el, false));
      return newElements;
    });
    return list;
  };
}

const SuspenseContext = createContext((delayMs = 0) => {
  let counter = 0,
      t,
      suspended = false;
  const [get, next] = createSignal(),
        store = {
    increment: () => {
      if (++counter === 1) {
        if (!store.initializing) {
          t = setTimeout(() => (suspended = true, next()), delayMs);
        } else suspended = true;
      }
    },
    decrement: () => {
      if (--counter === 0) {
        clearTimeout(t);

        if (suspended) {
          suspended = false;
          next();
        }
      }
    },
    suspended: () => {
      get();
      return suspended;
    },
    initializing: true
  };
  return store;
}); // lazy load a function component asynchronously

function lazy(fn) {
  return props => {
    const result = loadResource(fn().then(mod => mod.default));
    let Comp;
    return createMemo(() => (Comp = result.data) && sample(() => Comp(props)));
  };
} // load any async resource

function loadResource(resource) {
  const {
    increment,
    decrement
  } = useContext(SuspenseContext) || {};
  const [state, setState] = createState({
    loading: false
  });

  function doRequest(p, ref) {
    setState({
      loading: true
    });
    increment && increment();
    p.then(data => !(ref && ref.cancelled) && setState({
      data,
      loading: false
    })).catch(error => setState({
      error,
      loading: false
    })).finally(() => decrement && decrement());
  }

  if (typeof resource === "function") {
    createEffect(() => {
      let ref = {
        cancelled: false
      },
          res = resource();
      if (!res) return setState({
        data: undefined,
        loading: false
      });
      doRequest(res, ref);
      onCleanup(() => ref.cancelled = true);
    });
  } else doRequest(resource);

  return state;
}

exports.SuspenseContext = SuspenseContext;
exports.afterEffects = afterEffects;
exports.createContext = createContext;
exports.createDependentEffect = createDependentEffect;
exports.createEffect = createEffect;
exports.createMemo = createMemo;
exports.createRoot = createRoot;
exports.createSignal = createSignal;
exports.createState = createState;
exports.freeze = freeze;
exports.getContextOwner = getContextOwner;
exports.isListening = isListening;
exports.lazy = lazy;
exports.loadResource = loadResource;
exports.map = map;
exports.onCleanup = onCleanup;
exports.pipe = pipe;
exports.reconcile = reconcile;
exports.reduce = reduce;
exports.sample = sample;
exports.selectEach = selectEach;
exports.selectWhen = selectWhen;
exports.setDefaults = setDefaults;
exports.unwrap = unwrap;
exports.useContext = useContext;
